#include <errno.h>
#include <string.h>
#include <unistd.h>
#include "misc.h"
#include <lx_map.h>
#include "naw.h"
#include "naw_log.h"

#define MAX_PASSWORD_LENGTH 40

/***************************************************************************
  module: plain

  Plain text password authentication.  Simple module that can both act as an
  example and also provide some service isolation with minimum code exposure.
  Obviously, it does not provide any protection against MitM, replay, etc.

  For example, this module might be used in combination with x509
  authentication to avoid exposing openssl code publicly on the internet
  unnecessarily.

  Network passwords are supplied in the format:
    <len><string>

  where <len> is a single byte unsigned integer, and <string> is the supplied
  password, of length <len>.
 ***************************************************************************/


static inline int read_size(int sd, char *buf, size_t len) {
    ssize_t rlen;

    for (;;) {
        rlen = read(sd, buf, len);
        if (rlen == -1) {
            if (errno == EINTR)
                continue;
            naw_logfmt(NAW_NOTICE, "plain: read error: %s", strerror(errno));
            return NAW_ERR;
        }
        if (rlen < len) {
            buf += len;
            len -= rlen;
            continue;
        }
        break;
    }

    return NAW_OK;
}

static inline int write_size(int sd, char *buf, size_t len) {
    ssize_t rlen;

    for (;;) {
        rlen = write(sd, buf, len);
        if (rlen == -1) {
            if (errno == EINTR)
                continue;
            naw_logfmt(NAW_NOTICE, "plain: write error: %s", strerror(errno));
            return NAW_ERR;
        }
        if (rlen < len) {
            buf += len;
            len -= rlen;
            continue;
        }
        break;
    }

    return NAW_OK;
}

/***************************************************************************
  naw_test_module_accept

  
 ***************************************************************************/

static int naw_test_module_accept (handle, data)
    struct naw_handle *handle;
    struct naw_data *data;
{
    char *v;
    unsigned int vlen;
    unsigned int supplied_len;
    char c;

    if (!naw_data_get(data, "plainpw", &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("plain: plainpw: "),LXM_mb(v,vlen));
    } else {
        NAW_LOG(NAW_WARNING, LXM_q("plain: called with no plainpw: returning NO."));
        return NAW_NO;
    }

    if (vlen > MAX_PASSWORD_LENGTH) {
        NAW_LOG(NAW_WARNING, LXM_q("plain: plainpw > "), LXM_d(MAX_PASSWORD_LENGTH), LXM_q(" (truncating)"));
        vlen = MAX_PASSWORD_LENGTH;
    }

    /* read length from first byte */
    if (read_size(handle->sd, &c, 1) != NAW_OK)
        return NAW_NO;

    supplied_len = (unsigned int)c;
    if (supplied_len > MAX_PASSWORD_LENGTH) {
        NAW_LOG(NAW_NOTICE, LXM_q("plain: remote supplied length ("),
                LXM_du(supplied_len), LXM_q(") > max length ("),
                LXM_d(MAX_PASSWORD_LENGTH), LXM_q(")"));
        return NAW_NO;
    }

    /* read and compare the supplied password */
    {
        char supplied_pwbuf[supplied_len];

        if (read_size(handle->sd, supplied_pwbuf, supplied_len) != NAW_OK)
            return NAW_NO;

        /* XXX: may consider doing a full string comparison regardless of
         * length to avoid timing attacks, but it doesn't really matter that
         * much for this module */
        if (supplied_len != vlen || memcmp(v, supplied_pwbuf, vlen)) {
            NAW_LOG(NAW_NOTICE, LXM_q("plain: supplied password does not match"));
            return NAW_NO;
        }
    }

    NAW_LOG(NAW_DEBUG2, LXM_q("plain: accepting connection"));
    return NAW_YES;
}

static int naw_test_module_connect (handle, data)
    struct naw_handle *handle;
    struct naw_data *data;
{
    char *v;
    unsigned int vlen;

    if (!naw_data_get(data, "plainpw", &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("plain: plainpw: "),LXM_mb(v,vlen));
    } else {
        NAW_LOG(NAW_WARNING, LXM_q("plain: called with no plainpw: returning NO"));
        return NAW_NO;
    }

    {
        char wbuf[vlen + 1];
        wbuf[0] = (char)((ub1_t)vlen);
        memcpy(wbuf+1, v, vlen);
        if (write_size(handle->sd, wbuf, vlen+1) != NAW_OK) {
            naw_logfmt(NAW_NOTICE, "plain: write failed: %s", strerror(errno));
            /* XXX: don't need to explicitly fail the connection if this fails,
             * since the server should do it for us. */
            return NAW_NO;
        }

    }

    NAW_LOG(NAW_DEBUG2, LXM_q("plain: successful connection"));
    return NAW_YES;
}


struct naw_modfuncset naw_set = {
    naw_test_module_accept,  /* accept  */
    naw_test_module_connect, /* connect */
    NULL,                    /* init    */
    NULL,                    /* fini    */
};

